# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..io import IoManager
from ..model import Adapter, get_attribute_types, observe
from ..publish import IPublisher, ISubscriber
from ..ui import (Action, ActionCollection, Application, IApplication,
        IContainer, IDisplay, IView)

from .i_area_hints import IAreaHints
from .i_close_view_veto import ICloseViewVeto
from .i_open_model import IOpenModel
from .i_shell import IShell
from .i_tool import ITool


class BaseShellAdapter(Adapter):
    """ The BaseShellAdapter class is a base class for adapters to the IShell
    interface.  It implements housekeeping common to all shells.
    """

    def __init__(self):
        """ Initialise the adapter. """

        observe('active_view', IApplication(Application()),
                self.__active_view_changed)

        super(BaseShellAdapter, self).__init__()

    def close_view(self, view=None):
        """ Close a view if possible.

        :param view:
            is the view.  If it is not specified then the current view is
            closed.
        :return:
            ``True`` if the view was closed.
        """

        if view is None:
            view = self.current_view

        # Find the tool containing the view.
        for tool in self.tools:
            itool = ITool(tool)

            if view in itool.views:
                icloseviewveto = ICloseViewVeto(tool, exception=False)

                if icloseviewveto is None:
                    itool.views.remove(view)
                    self.views.remove(view)

                    return True

                return not icloseviewveto.veto(view)

        # We should never get here.
        return False

    def new_view_allowed(self):
        """ This determines if the main area policy allows for a new view to
        be added.

        :return:
            ``True`` if a new view is allowed.
        """

        # Handle the trivial case.
        if self.main_area_policy != 'single':
            return True

        # See if there is already a view in the main area.
        for view in self.views:
            if IAreaHints(view).area == '':
                # Try and close it.
                return self.close_view(view)

        return True

    def open(self, tool_id, location, format=''):
        """ Open the model at a location and add it to a tool.

        :param tool_id:
            is the identifier of the tool to handle the model.
        :param location:
            is the location as a string.
        :param format:
            is the identifier of the optional format.
        """

        # Get the tool.
        for tool in self.tools:
            if tool.id == tool_id:
                break
        else:
            raise ValueError(
                    "there is no tool with the identifier '{0}'".format(
                            tool_id))

        # Convert the location to a list of storage location instances.
        locations = IoManager.readable_locations_from_string(location, format)

        nr_locations = len(locations)

        if nr_locations == 0:
            Application.warning("Open",
                    "{0} is not a valid readable storage location.".format(
                            location), self.adaptee)
            return

        if nr_locations > 1:
            # Note that this is really a badly configured application.
            Application.warning("Open",
                    "{0} is an ambiguous readable storage location.".format(
                            location), self.adaptee)
            return

        # Ask the tool to open the model.
        IOpenModel(tool).open_model(locations[0])

    @IShell.publication_manager.default
    def publication_manager(self):
        """ Invoked to return the default publication manager. """

        from ..publish.default_publication_manager import PublicationManager

        return PublicationManager()

    @observe('tools')
    def __tools_changed(self, change):
        """ Invoked when the list of tools changes. """

        # Unbind the shell from any old tools.
        for tool in change.old:
            itool = ITool(tool)
            ishell = IShell(self.adaptee)
            iview = IView(self.adaptee)

            if IPublisher(tool, exception=False) is not None:
                self.publication_manager.publishers.remove(tool)

            if ISubscriber(tool, exception=False) is not None:
                self.publication_manager.subscribers.remove(tool)

            for action in itool.actions:
                iview.actions.remove(action)

            for view in itool.views:
                ishell.views.remove(view)

            itool.shell = None

        # Bind the shell to any new tools.
        for tool in change.new:
            itool = ITool(tool)
            ishell = IShell(self.adaptee)
            iview = IView(self.adaptee)

            itool.shell = self.adaptee

            for view in itool.views:
                ishell.views.append(view)

            if len(itool.actions) == 0:
                itool.actions = get_attribute_types(itool, ActionCollection)
                itool.actions.extend(get_attribute_types(itool, Action))

            iview.actions.extend(itool.actions)

            if ISubscriber(tool, exception=False) is not None:
                self.publication_manager.subscribers.append(tool)

            if IPublisher(tool, exception=False) is not None:
                self.publication_manager.publishers.append(tool)

    @observe('views')
    def __views_changed(self, change):
        """ Invoked when the list of views changes. """

        for view in change.old:
            observe('name', IDisplay(view), self.__view_name_changed,
                    remove=True)

        for view in change.new:
            observe('name', IDisplay(view), self.__view_name_changed)

    def __view_name_changed(self, change):
        """ Invoked when the name of a view changes. """

        self._update_window_title(change.new)

    def _update_window_title(self, view_name):
        """ Update the shell's window title when the name of a view changes.
        """

        if self.window_title_template != "":
            IView(self.adaptee).window_title = self.window_title_template.replace('[view]', view_name)

    def __active_view_changed(self, change):
        """ Invoked when the active view changes. """

        # Only update the current view if the active view is in one of the
        # shell's views.  This means that the current view will not change when
        # the application is not active.
        active_view = change.new

        if active_view is None:
            # However if there is no active view because our last view has been
            # closed then reset the current view.
            no_views = True
            for tool in self.tools:
                itool = ITool(tool)

                if len(itool.views) == 0:
                    itool.current_view = None
                else:
                    no_views = False

            if no_views:
                self.current_view = None
        else:
            for tool in self.tools:
                itool = ITool(tool)

                for view in itool.views:
                    if self._container_has_view(view, active_view):
                        # Reset the current tool of all other tools.
                        for reset_tool in self.tools:
                            if reset_tool is not tool:
                                ITool(reset_tool).current_view = None

                        itool.current_view = self.current_view = view

                        # Update the window title if there is a template.
                        self._update_window_title(IDisplay(view).name)

                        return

    @classmethod
    def _container_has_view(cls, container, view):
        """ See if a (possible) container contains a view. """

        if container is view:
            return True

        icontainer = IContainer(container, exception=False)
        if icontainer is None:
            return False

        for subview in icontainer.views:
            if cls._container_has_view(subview, view):
                return True

        return False
