# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ....io import (ICodec, IoManager, IIoManager, IIoManagerUi, IStorage,
        IStorageLocation, StorageError)
from ....model import implements, Instance, Model, observe, Str
from ....publish import Publication, IPublisher
from ....ui import IAction, Application, IDisplay, IView
from ....ui.actions import (CloseAction, NewAction, OpenAction, SaveAction,
        SaveAsAction)

from ... import (IDirty, IManagedModel, IManagedModelTool, IModelManagerTool,
        IQuitVeto, IShell)

from .resources_state import ResourcesState


@implements(IModelManagerTool, IDirty, IPublisher, IQuitVeto)
class ModelManagerTool(Model):
    """ The ModelManagerTool class is a :term:`tool` that handles the lifecycle
    of :term:`managed models<managed model>` in a :term:`shell`.
    """

    # The title of the wizard page to choose a model.
    choose_model_wizard_title = Str("Choose an item")

    # The text to use to prompt the user to choose a new model.
    # FIXME: The newlines are needed for OS/X (don't know about other
    # FIXME: platforms).  Should it be moved to the toolkit implementation?
    choose_new_model_prompt = Str("From the list below choose what sort of item you want to create.\n")

    # The text to use to prompt the user to choose a model to open.
    choose_open_model_prompt = Str("From the list below choose what sort of item you want to open.\n")

    # The text to use to prompt the user to choose a tool.
    choose_tool_prompt = Str("From the list below choose what tool you want to use.\n")

    # The title of the wizard page to choose a tool.
    choose_tool_wizard_title = Str("Choose a tool")

    # The Close action.
    close_action = CloseAction()

    # The tool's identifier.
    id = 'dip.shell.tools.model_manager'

    # The New action.
    new_action = NewAction()

    # The Open action.
    open_action = OpenAction()

    # The Save action.
    save_action = SaveAction()

    # The Save As action.
    save_as_action = SaveAsAction()

    # The text used as the question asking if the user really wants to close.
    user_close_question = Str("Do you still want to close?")

    # The text used as the title of the "Close" related views.
    user_close_title = Str()

    # The text used as the title of the "New" related views.
    user_new_title = Str()

    # The text used as the title of the "Open" related views.
    user_open_title = Str()

    # The text used as the title of the "Save" related views.
    user_save_title = Str()

    # The text used as the title of the "Save As" related views.
    user_save_as_title = Str()

    # The current model.
    _current_model = Instance(IManagedModel)

    # The current tool.
    _current_tool = Instance(IManagedModelTool)

    # The current default location to use when opening a model.
    _default_location = Instance(IStorageLocation)

    # The state of the different resources and their interrelationships.
    _resources = Instance(ResourcesState)

    def __init__(self):
        """ Initialise the tool. """

        super(ModelManagerTool, self).__init__()

        iiomanager = IIoManager(IoManager().instance)
        observe('codecs', iiomanager, lambda c: self._update_resources())
        observe('storage_factories', iiomanager,
                lambda c: self._update_resources())

    def veto_close_view(self, view, tool):
        """ This handles :meth:`~dip.shell.ICloseViewVeto.veto` on behalf of an
        implementation of the :class:`~dip.shell.IManagedModelTool` interface.

        :param view:
            is the view.
        :param tool:
            is the tool.
        :return:
            ``True`` if the close of the view is to be prevented.
        """

        imanagedmodeltool = IManagedModelTool(tool)

        model_to_remove = None

        # Find the model related to the view.
        for model in imanagedmodeltool.models:
            imanagedmodel = IManagedModel(model)
            if view in imanagedmodel.views:
                if imanagedmodel.dirty:
                    if imanagedmodel.invalid_reason != '':
                        answer = Application.question(self.user_close_title,
                                self._no_save_reason(model),
                                detail=imanagedmodel.invalid_reason,
                                parent=self.shell,
                                buttons=('cancel', 'discard'))

                        if answer == 'discard':
                            model_to_remove = model
                    else:
                        answer = Application.question(self.user_close_title,
                                self._no_close_reason(model),
                                parent=self.shell,
                                buttons=('save', 'cancel', 'discard'))

                        if answer == 'save':
                            # Do a Save or Save As.
                            if imanagedmodel.location is None:
                                self._save_model_as(model)
                            else:
                                self._save_model(model)

                            # If the model is no longer dirty then it was
                            # written successfully.
                            if not imanagedmodel.dirty:
                                model_to_remove = model
                        elif answer == 'discard':
                            model_to_remove = model
                else:
                    model_to_remove = model

                break

        if model_to_remove is None:
            return True

        imanagedmodeltool.models.remove(model)
        self._publish_close(model)

        return False

    def open_model(self, location, tool):
        """ This handles :meth:`~dip.shell.IOpenModel.open_model` on behalf of
        an implementation of the :class:`~dip.shell.IManagedModelTool`
        interface.

        :param location:
            is the storage location.
        :param tool:
            is the tool.
        """

        imanagedmodeltool = IManagedModelTool(tool)

        # There should be at least one codec and they all will have the same
        # format.
        format = ICodec(IStorage(IStorageLocation(location).storage).codecs[0]).format

        models = []
        abandoned = False
        for model_template in self._resources.model_templates():
            if not imanagedmodeltool.handles(model_template):
                continue

            model = self._resources.model_from_template(model_template)
            if IManagedModel(model).native_format != format:
                continue

            model = self._read_model(model, location)
            if model is None:
                abandoned = True
            else:
                models.append(model)

        nr_models = len(models)

        if nr_models == 1:
            model = models[0]
            self._publish_open(model)
            imanagedmodeltool.models.append(model)
        elif nr_models > 1:
            # Note that this is really a badly configured application.
            Application.warning(self.user_open_title,
                    "There are multiple types of item that can be read from "
                    "{0}".format(str(location)))
        elif not abandoned:
            # Only show a warning if there user hasn't already been involved.
            Application.warning(self.user_open_title,
                    "There is no type of item that can be read from "
                    "{0}".format(str(location)))

    @IQuitVeto.reasons.getter
    def reasons(self):
        """ Invoked to get the reasons why the application should not quit. """

        reasons = []

        for tool in self._managed_model_tools():
            for model in IManagedModelTool(tool).models:
                if IManagedModel(model).dirty:
                    reasons.append(self._no_close_reason(model))

        return reasons

    @user_close_title.default
    def user_close_title(self):
        """ Invoked to return the default user close title. """

        return IAction(self.close_action).plain_text

    @user_new_title.default
    def user_new_title(self):
        """ Invoked to return the default user new title. """

        return IAction(self.new_action).plain_text

    @user_open_title.default
    def user_open_title(self):
        """ Invoked to return the default user open title. """

        return IAction(self.open_action).plain_text

    @user_save_title.default
    def user_save_title(self):
        """ Invoked to return the default user save title. """

        return IAction(self.save_action).plain_text

    @user_save_as_title.default
    def user_save_as_title(self):
        """ Invoked to return the default user save as title. """

        return IAction(self.save_as_action).plain_text

    @observe('model_factories')
    def __model_factories_changed(self, change):
        """ Invoked when the list of model factories changes. """

        self._update_resources()

        if self.shell is not None:
            self._auto_new(change.new)

    def __shell_ready(self, change):
        """ Invoked when the shell is ready to be made visible. """

        self._auto_new()

    def _auto_new(self, model_factories=None):
        """ Add new models to any tool whose model policy is 'one' but don't
        yet have a model.  Also manage the visibility of the Close action.
        """

        if model_factories is None:
            model_factories = self.model_factories

        close_visible = False

        # Get the list of candidate tools.
        imanagedmodeltools = []
        for tool in self._managed_model_tools():
            imanagedmodeltool = IManagedModelTool(tool)

            if imanagedmodeltool.model_policy == 'one':
                if len(imanagedmodeltool.models) == 0:
                    imanagedmodeltools.append(imanagedmodeltool)
            else:
                close_visible = True

        IAction(self.close_action).visible = close_visible

        if len(imanagedmodeltools) == 0:
            return

        for model_factory in model_factories:
            model_template = self._resources.template_from_factory(
                    model_factory)

            for imanagedmodeltool in imanagedmodeltools:
                if imanagedmodeltool.handles(model_template):
                    model = self._resources.model_from_template(model_template)
                    self._publish_open(model)
                    imanagedmodeltool.models.append(model)

    def __tool_models_changed(self, change):
        """ Invoked when a tool's list of managed models changes. """

        for model in change.old:
            observe('dirty', IManagedModel(model), self.__model_dirty_changed,
                    remove=True)

        for model in change.new:
            observe('dirty', IManagedModel(model), self.__model_dirty_changed)

        self._update_dirty()

    def __model_dirty_changed(self, change):
        """ Invoked when a managed model's dirty state changes. """

        if change.new:
            self.dirty = True
        else:
            self._update_dirty()

    def _update_dirty(self):
        """ Update this tool's dirty state. """

        dirty = False

        for tool in self._managed_model_tools():
            for model in IManagedModelTool(tool).models:
                if IManagedModel(model).dirty:
                    dirty = True
                    break

        self.dirty = dirty

    def __tools_changed(self, change):
        """ Invoked when the list of tools changes. """

        for tool in change.old:
            imanagedmodeltool = IManagedModelTool(tool, exception=False)
            if imanagedmodeltool is not None:
                self._detach_tool(imanagedmodeltool)

        for tool in change.new:
            imanagedmodeltool = IManagedModelTool(tool, exception=False)
            if imanagedmodeltool is not None:
                self._attach_tool(imanagedmodeltool)

        self._update_resources()
        self._auto_new()

    def _attach_tool(self, imanagedmodeltool):
        """ Attach a managed model tool. """

        observe('models', imanagedmodeltool, self.__tool_models_changed)
        imanagedmodeltool.manager = self

    def _detach_tool(self, imanagedmodeltool):
        """ Detach a managed model tool. """

        imanagedmodeltool.manager = None
        observe('models', imanagedmodeltool, self.__tool_models_changed,
                remove=True)

    @observe('shell')
    def __shell_changed(self, change):
        """ Invoked when the shell changes. """

        # Remove from any old shell.
        if change.old is not None:
            for tool in self._managed_model_tools(change.old):
                self._detach_tool(IManagedModelTool(tool))

            ishell = IShell(change.old)

            observe('tools', ishell, self.__tools_changed, remove=True)
            observe('current_view', ishell, self.__current_view_changed,
                    remove=True)

            observe('ready', IView(change.old), self.__shell_ready,
                    remove=True)

        # Add to any new shell.
        if change.new is not None:
            for tool in self._managed_model_tools(change.new):
                self._attach_tool(IManagedModelTool(tool))

            self._update_resources()

            ishell = IShell(change.new)

            # Get the actions into a known state.
            iaction = IAction(self.close_action)
            iaction.enabled = False

            iaction = IAction(self.new_action)
            iaction.visible = self._resources.new_is_valid()

            iaction = IAction(self.open_action)
            iaction.visible = self._resources.open_is_valid()

            save_are_valid = self._resources.save_are_valid()

            iaction = IAction(self.save_action)
            iaction.visible = save_are_valid
            iaction.enabled = False

            iaction = IAction(self.save_as_action)
            iaction.visible = save_are_valid
            iaction.enabled = False

            observe('current_view', ishell, self.__current_view_changed)
            observe('tools', ishell, self.__tools_changed)

            observe('ready', IView(change.new), self.__shell_ready)

    @close_action.triggered
    def close_action(self):
        """ Invoked when the Close action is triggered. """

        IShell(self.shell).close_view()

    @new_action.triggered
    def new_action(self):
        """ Invoked when the New action is triggered. """

        model_template = self._resources.only_model_template_for_new()

        if model_template is not None:
            # Get the tools.
            tools = self._resources.tools_for_new(model_template)

            if len(tools) > 1:
                # Ask for the tool.

                from .tool_dialog import ToolDialog

                dialog = ToolDialog(model_manager=self, tools=tools)

                if not dialog.execute():
                    return

                tool = dialog.tool
            else:
                # We know the tool.
                tool = tools[0]

        elif self._resources.are_multiple_tools_for_new():
            # We need to ask the user for both the model factory and tool.

            from .new_wizard import NewWizard

            wizard = NewWizard(model_manager=self)

            if not wizard.execute():
                return

            model_template = wizard.model_template
            tool = wizard.tool

        else:
            # We will know the tool once we know the model template.

            from .model_template_dialog import ModelTemplateDialog

            dialog = ModelTemplateDialog(model_manager=self)

            if not dialog.execute():
                return

            model_template = dialog.model_template

            tools = self._resources.tools_for_new(model_template)
            tool = tools[0]

        if self._new_views_allowed(tool):
            # Create an instance of the model and add it to the tool.
            model = self._resources.model_from_template(model_template)
            self._publish_open(model)
            IManagedModelTool(tool).models.append(model)

    @open_action.triggered
    def open_action(self):
        """ Invoked when the Open action is triggered. """

        model_template, storage, tool = self._resources.only_model_template_for_open()

        if model_template is None or storage is None or tool is None:
            from .open_wizard import OpenWizard

            wizard = OpenWizard(model_manager=self,
                    location=self._default_location,
                    model_template=model_template, storage=storage, tool=tool)

            if not wizard.execute():
                return

            model_template = wizard.model_template
            location = wizard.location
            tool = wizard.tool

        else:
            # We know everything except the location.
            location = storage.ui.get_read_location(self.user_open_title,
                    default_location=self._default_location,
                    hints=model_template, parent=self.shell)

            if location is None:
                return

        # Remember for the next dialog/wizard.
        self._default_location = location

        if not self._new_views_allowed(tool):
            return

        # Create and read the model.
        model = self._resources.model_from_template(model_template)

        try:
            model = self._read_model(model, location)
        except StorageError as e:
            Application.warning("Read Error",
                    "There was an error reading {0}.".format(
                            IDisplay(model).name),
                    self.shell, str(e))

            return

        if model is not None:
            self._publish_open(model)
            IManagedModelTool(tool).models.append(model)

    @save_action.triggered
    def save_action(self):
        """ Invoked when the Save action is triggered. """

        if not self._saveable(self.user_save_title):
            return

        self._save_model(self._current_model)

    @save_as_action.triggered
    def save_as_action(self):
        """ Invoked when the Save As action is triggered. """

        if not self._saveable(self.user_save_as_title):
            return

        model = self._current_model

        self._save_model_as(model)
        IAction(self.save_action).enabled = (IManagedModel(model).location is not None)

    def _save_model_as(self, model):
        """ Save a model at a new location. """

        imanagedmodel = IManagedModel(model)

        default_location = imanagedmodel.location
        if default_location is None:
            default_location = self._default_location

        location = IIoManagerUi(IoManager.ui).writeable_storage_location(
                self.user_save_as_title, default_location=default_location,
                format=imanagedmodel.native_format, hints=model,
                parent=self.shell)

        if location is None:
            return

        self._default_location = imanagedmodel.location = location

        self._save_model(model)

    def __current_view_changed(self, change):
        """ Invoked when the shell's current view changes. """

        self._current_tool, self._current_model = self._tool_model_from_view(
                change.new)

        enabled = (self._current_model is not None)

        if enabled:
            close_enabled = (IManagedModelTool(self._current_tool).model_policy != 'one')

            location = IManagedModel(self._current_model).location
            save_enabled = (location is not None)
        else:
            close_enabled = False
            save_enabled = False

        IAction(self.close_action).enabled = close_enabled
        IAction(self.save_action).enabled = save_enabled
        IAction(self.save_as_action).enabled = enabled

    def _new_views_allowed(self, tool):
        """ Try and get things into a state where a new view will be allowed.
        """

        imanagedmodeltool = IManagedModelTool(tool)

        if imanagedmodeltool.model_policy != 'many':
            return IShell(self.shell).close_view(imanagedmodeltool.views[0])

        return IShell(self.shell).new_view_allowed()

    def _tool_model_from_view(self, view):
        """ Return the managed tool and model being handled by a view, if any.
        """

        if view is not None:
            for tool in self._managed_model_tools():
                for model in IManagedModelTool(tool).models:
                    if view in IManagedModel(model).views:
                        return tool, model

        return None, None

    @_resources.default
    def _resources(self):
        """ Invoked to return the default resources state. """

        return ResourcesState()

    def _update_resources(self):
        """ Update the internal state when the resources have changed. """

        if self.shell is None:
            return

        self._resources.update(self.model_factories, IShell(self.shell).tools)

        # Update the state of the dependent actions.
        IAction(self.new_action).visible = self._resources.new_is_valid()
        IAction(self.open_action).visible = self._resources.open_is_valid()
        IAction(self.save_action).visible = IAction(self.save_as_action).visible = self._resources.save_are_valid()

    def _saveable(self, title):
        """ Tell the use if the current model cannot be saved. """

        imanagedmodel = IManagedModel(self._current_model)

        if imanagedmodel.invalid_reason == '':
            return True

        # Note that we don't tie the enabled state of the action to the invalid
        # reason.  This is so that we get the opportunity to tell the user why
        # the model cannot be saved.

        Application.warning(title, self._no_save_reason(self._current_model),
                detail=imanagedmodel.invalid_reason, parent=self.shell)

        return False

    def _save_model(self, model):
        """ Save a model. """

        imanagedmodel = IManagedModel(model)
        location = imanagedmodel.location

        try:
            location.storage.write(model, location)
        except StorageError as e:
            Application.warning("Write Error",
                    "There was an error writing {0}.".format(
                            IDisplay(model).name),
                    self.shell, str(e))

            return

        imanagedmodel.dirty = False

    def _managed_model_tools(self, shell=None):
        """ Return the list of managed model tools. """

        if shell is None:
            shell = self.shell

        return [tool for tool in IShell(shell).tools
                if IManagedModelTool(tool, exception=False) is not None]

    @staticmethod
    def _no_close_reason(model):
        """ Return the reason why a managed model should not be closed. """

        return "{0} has unsaved changes.".format(IDisplay(model).name)

    @staticmethod
    def _no_save_reason(model):
        """ Return the reason why a managed model cannot be closed. """

        return "{0} cannot be saved in its current state.".format(
                IDisplay(model).name)

    def _read_model(self, model, location):
        """ Read a model from a location. """

        # Add the management data now so that the codec can update the dirty
        # flag if it wants to.
        imanagedmodel = IManagedModel(model)
        imanagedmodel.location = location
        imanagedmodel.dirty = False

        read_model = location.storage.read(model, location)

        if read_model is not None and read_model is not model:
            # Add the management data if the codec created its own model.
            imm = IManagedModel(read_model)
            imm.location = imanagedmodel.location
            imm.dirty = imanagedmodel.dirty

        return read_model

    def _publish_open(self, model):
        """ Publish the opening of a model. """

        self.publication = Publication(model=model, event='dip.events.opened')

    def _publish_close(self, model):
        """ Publish the closing of a model. """

        self.publication = Publication(model=model, event='dip.events.closed')
