# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from PyQt4.QtCore import QEvent, QObject
from PyQt4.QtGui import QFormLayout, QLayout

from ....model import adapt, Instance
from ....ui import IForm, IView

from ..layout_adapter import LayoutAdapter


@adapt(QFormLayout, to=IForm)
class QFormLayoutIFormAdapter(LayoutAdapter):
    """ An adapter to implement IForm for a QFormLayout. """

    # The field visibility event filter.
    _visibility_event_filter = Instance(QObject)

    @IForm.views.getter
    def views(self):
        """ Invoked to get the form's views. """

        layout = self.adaptee

        views = []
        for row in range(layout.rowCount()):
            itm = layout.itemAt(row, QFormLayout.FieldRole)
            if itm is None:
                itm = layout.itemAt(row, QFormLayout.SpanningRole)

            if itm is None:
                w = None
            else:
                w = itm.widget()
                if w is None:
                    w = itm.layout()

            views.append(w)

        return tuple(views)

    @views.setter
    def views(self, value):
        """ Invoked to set the form's views. """

        layout = self.adaptee

        # Clear the layout.
        while layout.count() > 0:
            layout.takeAt(0)

        # Populate the layout.
        for row, field in enumerate(value):
            try:
                label = self.labels[row]
            except IndexError:
                label = None

            if label is None:
                if isinstance(field, QLayout):
                    layout.setLayout(row, QFormLayout.SpanningRole, field)
                else:
                    layout.setWidget(row, QFormLayout.SpanningRole, field)
            else:
                layout.addRow(label, field)

                # Set the label's enabled and visible states to match the
                # field's.  Note that Qt won't actually create a label widget
                # if the label is an empty string.
                label_w = layout.labelForField(field)
                if label_w is not None:
                    if isinstance(field, QLayout):
                        iview = IView(field)
                        enabled = iview.enabled
                        visible = iview.visible
                    else:
                        # Install an event filter for the field to monitor its
                        # visibility.
                        field.installEventFilter(self._visibility_event_filter)

                        enabled = field.isEnabled()
                        visible = field.isVisible()

                    label_w.setEnabled(enabled)
                    label_w.setVisible(visible)

        self.qt4_configure(value)

    @_visibility_event_filter.default
    def _visibility_event_filter(self):
        """ Invoked to return the default field visibility event filter. """

        return VisibilityEventFilter(self.adaptee)


class VisibilityEventFilter(QObject):
    """ VisibilityEventFilter is an internal class that monitors the visibility
    of field widgets.
    """

    def eventFilter(self, obj, event):
        """ Reimplemented to handle any hide and show events. """

        event_type = event.type()

        if event_type == QEvent.Hide:
            self.parent().labelForField(obj).setVisible(False)
        elif event_type == QEvent.Show:
            label = self.parent().labelForField(obj)
            # FIXME: This doesn't seem to come into effect until the window is
            #        activated.
            label.setVisible(True)

            # We also make sure the label's enabled state is up to date.  We
            # can't rely on the setter in the widget adapter because the field
            # widget's layout manager may not have been set at that time.  We
            # also can't rely on the field's state because it may have been set
            # by Qt (e.g. in response to changing the state of a parent), so we
            # use the shadow attribute which reflects the programmer's intent.
            label.setEnabled(IView(obj)._enabled)

        return super(VisibilityEventFilter, self).eventFilter(obj, event)
