# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import (implements, Instance, List, Model, observe,
        ValueTypeFactory)

from .i_action import IAction
from .i_action_binding import IActionBinding
from .i_container import IContainer
from .i_controller import IController
from .i_editor import IEditor
from .i_editor_binding import IEditorBinding
from .i_message_area import IMessageArea
from .i_validator import IValidator
from .i_view import IView


@implements(IController)
class Controller(Model):
    """ The Controller class is the default implementation of a
    :term:`controller`.
    """

    # The editor that is currently causing a change to the model.
    _changing_editor = Instance(IEditor)

    # The list of invalid editors.
    _invalid_editors = List(IEditor)

    def add_action(self, action):
        """ Add an action to the controller.

        :param action:
            is the action.
        """

        self.actions.append(action)

        # There is nothing else to do if the action isn't bound.
        binding = IActionBinding(action)
        if binding.model is None:
            return

        iaction = IAction(action)

        # If the action is bound to an attribute then treat it like an editor.
        if binding.attribute_is_type:
            model = binding.model
            name = binding.attribute_name

            # Get the initial checked state.
            if iaction.checkable:
                iaction.checked = getattr(model, name)

            # Watch for any subsequent model changes.
            observe(name, model,
                    lambda c: self._action_attr_changed(action, c.new))

        # Handle any action changes.
        iaction.when_triggered = self._action_triggered

    def add_editor(self, editor):
        """ Add an editor to the controller.

        :param editor:
            is the editor.
        """

        self.editors.append(editor)

        # Do any setup that is specific to types that represent a value.
        ieditor = IEditor(editor)
        binding = IEditorBinding(editor)

        if isinstance(binding.attribute_type, ValueTypeFactory):
            model = binding.model
            name = binding.attribute_name

            # Set the initial value.
            ieditor.value = getattr(model, name)

            # Watch for any subsequent model changes.
            observe(name, model,
                    lambda c: self._editor_attr_changed(editor, c.new))

        # Watch for any view changes.
        observe('value', ieditor,
                lambda c: self._editor_changed(editor, c.new))

    def update_model(self):
        """ Update the all the attributes of a model that have a corresponding
        editor.
        """

        for editor in self.editors_to_validate():
            if self._valid_editor(editor):
                binding = IEditorBinding(editor)

                setattr(binding.model, binding.attribute_name,
                        IEditor(editor).value)

        self.update_view()

    def validate(self):
        """ Validate the view and update the explanation as to why it is (or
        isn't) invalid.
        """

        self._invalid_editors = []

        invalid_reason = ''

        for editor in self.editors_to_validate():
            if self._valid_editor(editor):
                ieditor = IEditor(editor)

                if ieditor.validator is not None:
                    reason = IValidator(ieditor.validator).validate(self,
                            editor, ieditor.value)

                    if reason != '':
                        self._invalid_editors.append(editor)

                        if invalid_reason == '':
                            # Remember the first reason.
                            invalid_reason = "{0}: {1}.".format(ieditor.label,
                                    reason)

        if invalid_reason == '':
            invalid_reason = self.validate_view()

        for message_area in self.message_areas:
            IMessageArea(message_area).text = invalid_reason

        self.invalid_reason = invalid_reason

    def editors_to_validate(self):
        """ Return the sequence of editors to validate.

        :return:
            the sequence of editors.
        """

        return self.editors

    def validate_view(self):
        """ This is called after the view has been updated and only if all the
        view's editors contain valid values giving the controller an
        opportunity to carry out cross-editor validation.

        :return:
            a string which will be empty if the view is valid, otherwise it
            will explain why the view is invalid.  This implementation always
            returns an empty string.
        """

        return ''

    def update_view(self):
        """ This is called after the view has been validated giving the
        controller an opportunity to update the state of the view.  This
        implementation does nothing.
        """

    def is_valid(self, editor=None):
        """ Return the validity of an editor's (or all editors) current value.

        :param editor:
            is the editor.  If this is ``None`` then the validity of all
            editors is returned.
        :return:
            ``True`` if the editor (or all editors) is valid.
        """

        if editor is None:
            return (self.invalid_reason == '')

        return (editor not in self._invalid_editors)

    def find_action(self, id, tail=False):
        """ Find the action with a particular identifier.

        :param id:
            is the identifier of the action.
        :param tail:
            if this is set then only the last part of a dotted action
            identifier is compared.
        :return:
            the action or ``None`` if there was no such action.
        """

        for action in self.actions:
            target = IAction(action).id
            if tail:
                target = target.split('.')[-1]

            if target == id:
                return action

        return None

    def find_editor(self, id, tail=False):
        """ Find the editor with a particular identifier.

        :param id:
            is the identifier of the editor.
        :param tail:
            if this is set then only the last part of a dotted editor
            identifier is compared.
        :return:
            the editor or ``None`` if there was no such editor.
        """

        for editor in self.editors:
            target = IEditor(editor).id
            if tail:
                target = target.split('.')[-1]

            if target == id:
                return editor

        return None

    def __getattr__(self, name):
        """ Reimplemented to support convenient access to actions and editors
        using a name derived from an identifier.

        :param name:
            the name of the attribute.  If this has a suffix of ``_action``,
            ``_editor`` or ``_view`` then the rest of the name is interpreted
            as the identifier of an action or editor (which is normally the
            name of the model attribute that the action or editor is bound to)
            or view.
        :return:
            the attribute.  An AttributeError will be raised if there was no
            such attribute, action or editor.
        """

        id = self._extract_id(name, '_action')
        if id != '':
            action = self.find_action(id, tail=True)
            if action is not None:
                return action

        else:
            id = self._extract_id(name, '_editor')
            if id != '':
                editor = self.find_editor(id, tail=True)
                if editor is not None:
                    return editor

            else:
                id = self._extract_id(name, '_view')
                if id != '':
                    for view in self._get_views(self.view):
                        if IView(view).id.split('.')[-1] == id:
                            return view

        raise AttributeError(
                "'{0}' object has no attribute '{1}'".format(
                        type(self).__name__, name))

    @classmethod
    def _get_views(cls, root):
        """ A generator to return all the views starting with a given root. """

        yield root

        icontainer = IContainer(root, exception=False)
        if icontainer is not None:
            for view in icontainer.views:
                # Not everything in a container is a view (e.g. Stretch()).
                if IView(view, exception=False) is None:
                    continue

                for child in cls._get_views(view):
                    yield child

    @staticmethod
    def _extract_id(name, suffix):
        """ Return the identifier from a suffixed attribute name. """

        return name[:-len(suffix)] if name.endswith(suffix) else ''

    @staticmethod
    def _valid_editor(editor):
        """ Return ``True`` if the editor is valid because we are interested in
        its value.
        """

        # Ignore disabled editors as they may legitimately contain invalid
        # values.
        return isinstance(IEditorBinding(editor).attribute_type, ValueTypeFactory) and IEditor(editor).enabled

    def _action_triggered(self, action):
        """ This is invoked when the user triggers an action. """

        binding = IActionBinding(action)

        # If the action is bound to an attribute then treat it like an editor.
        if binding.attribute_is_type:
            self.validate()

            if self.auto_update_model:
                # Prevent recursive changes.
                self._changing_action = action
                setattr(binding.model, binding.attribute_name,
                        IAction(action).checked)
                self._changing_action = None

            self.update_view()
        else:
            # The attribute should be a callable.
            getattr(binding.model, binding.attribute_name)(self)

    def _editor_changed(self, editor, value):
        """ This is invoked when the user changes the value of an editor. """

        binding = IEditorBinding(editor)

        if isinstance(binding.attribute_type, ValueTypeFactory):
            self.validate()

            if self.auto_update_model and self.is_valid(editor):
                # Prevent recursive changes.
                self._changing_editor = editor
                setattr(binding.model, binding.attribute_name, value)
                self._changing_editor = None

            self.update_view()
        else:
            # Non-value types must update the model immediately.
            setattr(binding.model, binding.attribute_name, value)

    def _action_attr_changed(self, action, value):
        """ Invoked when the attribute bound to an action changes. """

        # Don't do anything if this action triggered the change.
        if self._changing_action is not action:
            IAction(action).value = value

            self.validate()
            self.update_view()

    def _editor_attr_changed(self, editor, value):
        """ Invoked when the attribute bound to an editor changes. """

        # Don't do anything if this editor triggered the change.
        if self._changing_editor is not editor:
            IEditor(editor).value = value

            self.validate()
            self.update_view()
