# Copyright (c) 2011 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import (Bool, get_attribute_type, Instance,
        resolve_attribute_path, Str)

from .i_controller import IController
from .i_editor import IEditor
from .i_editor_binding import IEditorBinding
from .i_validator import IValidator
from .view_factory import ViewFactory


class EditorFactory(ViewFactory):
    """ The EditorFactory class is a factory for views that implement the
    :class:`~dip.ui.IEditor` interface.  The
    :attr:`~dip.ui.IView.status_tip`, :attr:`~dip.ui.IView.tool_tip` and
    :attr:`~dip.ui.IView.whats_this` attributes will default to any values
    defined in the type's meta-data.
    """

    # The name of the attribute that the editor is bound to.  This may be an
    # :term:`attribute path`.
    bind_to = Str()

    # A short, user friendly label for the attribute.  It is typically used as
    # a field name or the text of a push button.  If not given then any
    # corresponding value of the type's meta-data is used instead.  If that
    # hasn't been defined then a label based on the name of the attribute in
    # its containing model is used.
    label = Str(None, allow_none=True)

    # This is set if the editor is read-only.
    read_only = Bool(False)

    # The editor's validator.
    validator = Instance(IValidator)

    def __init__(self, bind_to='', **properties):
        """ Initialise the factory.

        :param bind_to:
            the name of the attribute that the editor is bound to.  It should
            only be an empty string if the factory is being used to create a
            delegate.
        :param \*\*properties:
            are keyword arguments used as property names and values that are
            applied to the editor.
        """

        super(EditorFactory, self).__init__(**properties)

        # It's more natural to allow this as a positional argument.
        self.bind_to = bind_to

    def create(self, model, controller, parent):
        """ Create a view instance.

        :param model:
            is the model.
        :param controller:
            is the controller.
        :param parent:
            is the optional parent view.
        :return:
            the view.
        """

        editor = self.object_for_interface(self.view_interface, parent)
        self.bind_editor(editor, model, self.bind_to, controller)

        return editor

    def bind_editor(self, editor, model, name, controller):
        """ Bind an editor to a model attribute and add it to a controller.

        :param editor:
            is the editor.
        :param model:
            is the model.
        :param name:
            is the name of the attribute.  It may be an :term:`attribute path`.
        :param controller:
            is the controller.
        """

        binding = IEditorBinding(editor)

        # It may not be bound if it is being used as a delegate.
        if name != '':
            name, model = resolve_attribute_path(name, model)
            binding.attribute_type = get_attribute_type(model, name)

        binding.attribute_name = name
        binding.model = model

        self.bind_view(editor, controller)

        IController(controller).add_editor(editor)

    def configure(self, view):
        """ Reimplemented to configure the IEditor specific attributes.

        :param view:
            is the view to configure.
        """

        ieditor = IEditor(view)

        # Install any validator.  Do it now to allow
        # configure_with_attribute_type() to make further changes.
        ieditor.validator = self.validator

        # Do the configuration that can use values from the bound attribute's
        # type.  Allow for the editor not being bound.
        attribute_type = IEditorBinding(view).attribute_type
        if attribute_type is not None:
            self.configure_with_attribute_type(view, attribute_type)

            # The label requires extra handling.
            label = self.label
            if label is None:
                label = attribute_type.metadata.get('label')

                if label is None:
                    if attribute_type.name is None:
                        label = ""
                    else:
                        label = self.get_natural_name(attribute_type.name)

            ieditor.label = label

        ieditor.read_only = self.read_only

        # Configure any validator.
        if ieditor.validator is not None:
            IValidator(ieditor.validator).configure(view)

        super(EditorFactory, self).configure(view)

    def configure_with_attribute_type(self, view, attribute_type):
        """ Configure the view possibly using values from the bound attribute's
        type.

        :param view:
            is the view to configure.
        :param attribute_type:
            is the attribute's type.
        """

        ieditor = IEditor(view)
        metadata = attribute_type.metadata

        status_tip = self.status_tip
        if status_tip is None:
            status_tip = metadata.get('status_tip')

        if status_tip is not None:
            ieditor.status_tip = status_tip

        tool_tip = self.tool_tip
        if tool_tip is None:
            tool_tip = metadata.get('tool_tip')

        if tool_tip is not None:
            ieditor.tool_tip = tool_tip

        whats_this = self.whats_this
        if whats_this is None:
            whats_this = metadata.get('whats_this')

        if whats_this is not None:
            ieditor.whats_this = whats_this

    @ViewFactory.id.default
    def id(self):
        """ The default value of id is the name of the attribute that the
        editor is bound to.  The identifier only needs to be explicitly set if
        there is more than one editor in the same view that is bound to the
        same attribute and the editor is the target of a linked editor.
        """

        return self.bind_to
