# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Enum

from .editor_factory import EditorFactory
from .i_line_editor import ILineEditor


class LineEditor(EditorFactory):
    """ The LineEditor class implements a line editor factory for string
    attributes.
    """

    # This specifies whether or not a value is required.  It is applied after
    # any validator has determined that the value is valid.  'no' means that a
    # value is not required (i.e. it can be an empty string).  'yes' means that
    # a value is required.  'stripped' means that a value with no leading or
    # trailing whitespace is required.  If not given (and a validator hasn't
    # been explicitly specified) then any corresponding value in the type's
    # meta-data is used instead.
    required = Enum('no', 'yes', 'stripped', allow_none=True)

    # The specific interface implemented by the view.
    view_interface = ILineEditor

    def configure_with_attribute_type(self, view, attribute_type):
        """ Configure the view possibly using values from the bound attribute's
        type.

        :param view:
            is the view to configure.
        :param attribute_type:
            is the attribute's type.
        """

        ilineeditor = ILineEditor(view)

        # See if the metadata implies a validator, but only if the factory
        # hasn't had one created explicitly.
        if ilineeditor.validator is None and self.required is None:
            # A value wasn't explicitly specified so check the metadata.
            required = attribute_type.metadata.get('required')

            if required is not None:
                # Create a validator specific to the metadata value.
                ilineeditor.validator = self._create_validator(required)

        super(LineEditor, self).configure_with_attribute_type(view,
                attribute_type)

    @EditorFactory.validator.default
    def validator(self):
        """ Invoked to return the default validator. """

        return self._create_validator(self.required)

    @staticmethod
    def _create_validator(required):
        """ Create an appropriate validator. """

        if required in ('yes', 'stripped'):
            from .string_validator import StringValidator

            validator = StringValidator(stripped=(required == 'stripped'))
        else:
            validator = None

        return validator
