# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Bool, Enum, implements, Model, Str
from ..io import IoManager

from .i_validator import IValidator


@implements(IValidator)
class StorageLocationValidator(Model):
    """ The StorageLocationValidator class implements an editor validator for
    storage locations.
    """

    # The identifier of the format.
    format = Str()

    # The mode of the editor.  'open' means an existing location will be
    # obtained.  'save' means that a location that may or may not exist will
    # be obtained.
    mode = Enum('open', 'save')

    # This is set if a value is required.
    required = Bool(False)

    def configure(self, editor):
        """ Configure an editor to use this validator.

        :param editor:
            is the editor.
        """

    def validate(self, controller, editor, value):
        """ Validate an editor's value.

        :param controller:
            is the controller.
        :param editor:
            is the editor.
        :param value:
            is the value.
        :return:
            a string describing why the value is invalid.  An empty string is
            returned if the value is valid.
        """

        if value == '':
            invalid_reason = "a storage location is required" if self.required else ''
        else:
            if self.mode == 'open':
                locations = IoManager.readable_locations_from_string(value,
                        self.format)
            else:
                locations = IoManager.writeable_locations_from_string(value,
                        self.format)

            nr_locations = len(locations)
            if nr_locations == 0:
                invalid_reason = "a valid {0} storage location is required".format("readable" if self.mode == 'open' else "writeable")
            elif nr_locations > 1:
                invalid_reason = "an unambiguous storage location is required"
            else:
                invalid_reason = ''

        return invalid_reason
