# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import (Any, Callable, get_model_types, Instance, List, MetaModel,
        Str)
from ..toolkits import Toolkit

from .i_editor_binding import IEditorBinding
from .editor_factory import EditorFactory
from .table_column import TableColumn
from .i_table_editor import ITableEditor


class TableEditor(EditorFactory):
    """ The TableEditor class implements a table editor factory for list
    attributes.
    """

    # The list of columns.  If a string is given then it is used as the label
    # (i.e. heading) of the column.
    columns = List(Instance(TableColumn, str))

    # When called this will create the data for a new row in the table.
    row_factory = Callable()

    # The specific interface implemented by the view.
    view_interface = ITableEditor

    def configure(self, view):
        """ Reimplemented to configure the ITableEditor specific attributes.

        :param view:
            is the view to configure.
        """

        itableedit = ITableEditor(view)

        if len(self.columns) == 0:
            raise AttributeError("no columns have been specified")

        # Create complete column specifications.
        columns = []

        # See what we have a list of.
        row_type = IEditorBinding(view).attribute_type.element_type
        row_factory = self.row_factory

        if not isinstance(row_type, List):
            raise TypeError("the elements of the List must be another List")

        element_type = row_type.element_type

        for col in self.columns:
            column_type = element_type

            if isinstance(col, str):
                column = TableColumn(label=col)
            else:
                if element_type is None:
                    column_type = col.column_type
                elif col.column_type is not None:
                    raise AttributeError(
                            "'column_type' should not be specified because it "
                            "can be inferred")

                column = TableColumn(editable=col.editable,
                        editor_factory=col.editor_factory, label=col.label)

            # Make sure there is a column type and an appropriate editor
            # factory.
            if column_type is None or isinstance(column_type, Any):
                column_type = Str()

            column.column_type = column_type

            if column.editor_factory is None:
                column.editor_factory = Toolkit.editor_factory_for_attribute_type(column_type)

            columns.append(column)

        if row_factory is None:
            if element_type is None:
                element_type = Str()

            element = element_type.get_default_value()
            nr_elements = len(columns)
            row_factory = lambda: [element] * nr_elements

        itableedit.columns = columns
        itableedit.row_factory = row_factory

        super(TableEditor, self).configure(view)

    @EditorFactory.validator.default
    def validator(self):
        """ Invoked to return the default validator. """

        from .collection_validator import CollectionValidator

        return CollectionValidator()
