# Copyright (c) 2010 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import Dict, Int, Model, observe, Str, ValidationError


class Klass(Model):
    """ A class for test models. """

    default_dict = Dict()

    explicit_default_dict = Dict(default={1: 'one'})

    attr = Dict(Str(), Int())

    def __init__(self):
        super(Klass, self).__init__()

        self.change = None
        self.changes = 0
        observe('attr', self, self.__observed)

    def __observed(self, change):
        self.change = change
        self.changes += 1


class TestDictType(unittest.TestCase):
    """ This tests the Dict type. """

    def check_observed(self, model, new, old):
        """ Check that a test model observed a change. """

        self.assertEquals(model.changes, 1)

        change = model.change

        self.assertIs(change.model, model)
        self.assertEquals(change.name, 'attr')
        self.assertEquals(change.new, new)
        self.assertEquals(change.old, old)

    def check_unobserved(self, model):
        """ Check that a test model hasn't observed a change. """

        self.assertEquals(model.changes, 0)

    def test_invalid_key_type(self):
        """ Invalid Dict key type. """

        class BadDict(Model):
            bad = Dict(0)

        with self.assertRaises(TypeError):
            BadDict(bad={})

    def test_invalid_value_type(self):
        """ Invalid Dict value type. """

        class BadDict(Model):
            bad = Dict(Str(), 0)

        with self.assertRaises(TypeError):
            BadDict(bad={})

    def test_default_dict_type_explicit_default(self):
        """ Default Dict explicit default value. """

        k = Klass()
        self.assertDictEqual(k.explicit_default_dict, {1: 'one'})

    def test_dict_subclass_default(self):
        """ Default Dict sub-class. """

        class SubKlass(Klass):

            default_dict = {1: 'one'}

        k = SubKlass()
        self.assertDictEqual(k.default_dict, {1: 'one'})

    def test_default_dict_type_default_value(self):
        """ Default Dict type default value. """

        k = Klass()
        self.assertDictEqual(k.default_dict, {})

    def test_default_dict_type_valid_value(self):
        """ Default Dict type valid value. """

        k = Klass()
        k.default_dict = {1: 'one', 'two': 2}
        self.assertDictEqual(k.default_dict, {1: 'one', 'two': 2})

    def test_default_dict_type_invalid_value(self):
        """ Default Dict type invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.default_dict = None

    def test_valid_dict_valid_default_value(self):
        """ Dict type valid default value. """

        k = Klass(attr={'abc': 10})
        self.assertDictEqual(k.attr, {'abc': 10})

    def test_valid_dict_invalid_default_value(self):
        """ Dict type invalid default value. """

        with self.assertRaises(ValidationError):
            Klass(attr={10: 'abc'})

    def test_valid_dict_rebind_valid_value(self):
        """ Dict type rebind valid value. """

        k = Klass()
        k.attr = {'abc': 10}
        self.assertDictEqual(k.attr, {'abc': 10})
        self.check_observed(k, {'abc': 10}, {})

    def test_valid_dict_rebind_unchanged_value(self):
        """ Dict type rebind unchangedvalid value. """

        k = Klass()
        k.attr = {}
        self.assertDictEqual(k.attr, {})
        self.check_unobserved(k)

    def test_valid_dict_rebind_invalid_value(self):
        """ Dict type rebind invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = {10: 'abc'}

    def test_valid_dict_rebind_invalid_dict(self):
        """ Dict type rebind invalid dict. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = k.default_dict

    def test_dict_clear(self):
        """ Dict clear(). """

        k = Klass(attr={'abc': 10})
        k.attr.clear()
        self.assertDictEqual(k.attr, {})
        self.check_observed(k, {}, {'abc': 10})

    def test_dict_clear_unchanged(self):
        """ Unchanged dict clear(). """

        k = Klass()
        k.attr.clear()
        self.assertDictEqual(k.attr, {})
        self.check_unobserved(k)

    def test_dict_pop(self):
        """ Dict pop(). """

        k = Klass(attr={'abc': 10})
        value = k.attr.pop('abc')
        self.assertEqual(value, 10)
        self.assertDictEqual(k.attr, {})
        self.check_observed(k, {}, {'abc': 10})

    def test_dict_pop_default(self):
        """ Dict pop() with a default value. """

        k = Klass(attr={'abc': 10})
        value = k.attr.pop('xyz', 20)
        self.assertEqual(value, 20)
        self.assertDictEqual(k.attr, {'abc': 10})
        self.check_unobserved(k)

    def test_dict_popitem(self):
        """ Dict popitem(). """

        k = Klass(attr={'abc': 10})
        key, value = k.attr.popitem()
        self.assertEqual(key, 'abc')
        self.assertEqual(value, 10)
        self.assertDictEqual(k.attr, {})
        self.check_observed(k, {}, {'abc': 10})

    def test_dict_setdefault_valid(self):
        """ Dict setdefault() valid key/value. """

        k = Klass()
        value = k.attr.setdefault('abc', 10)
        self.assertEqual(value, 10)
        self.assertDictEqual(k.attr, {'abc': 10})
        self.check_observed(k, {'abc': 10}, {})

    def test_dict_setdefault_unchanged(self):
        """ Unchanged dict setdefault() valid key/value. """

        k = Klass(attr={'abc': 10})
        value = k.attr.setdefault('abc', 20)
        self.assertEqual(value, 10)
        self.assertDictEqual(k.attr, {'abc': 10})
        self.check_unobserved(k)

    def test_dict_setdefault_invalid_key(self):
        """ Dict setdefault() invalid key. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr.setdefault(10)

    def test_dict_setdefault_invalid_value(self):
        """ Dict setdefault() invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr.setdefault('abc')

    def test_dict_update_valid(self):
        """ Dict update() valid value. """

        k = Klass(attr={'abc': 20})
        k.attr.update({'abc': 10, 'xyz': 100})
        self.assertDictEqual(k.attr, {'abc': 10, 'xyz': 100})
        self.check_observed(k, {'abc': 10, 'xyz': 100}, {'abc': 20})

    def test_dict_update_unchanged(self):
        """ Unchanged dict update() valid value. """

        k = Klass(attr={'abc': 10})
        k.attr.update({'abc': 10})
        self.assertDictEqual(k.attr, {'abc': 10})
        self.check_unobserved(k)

    def test_dict_update_invalid(self):
        """ Dict update() invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr.update({10: 'abc'})

    def test_dict_indexing_valid(self):
        """ Dict [] valid key/value. """

        k = Klass(attr={'abc': 10})
        k.attr['abc'] = 20
        self.assertDictEqual(k.attr, {'abc': 20})
        self.check_observed(k, {'abc': 20}, {'abc': 10})

    def test_dict_indexing_unchanged(self):
        """ Unchanged dict [] valid key/value. """

        k = Klass(attr={'abc': 10})
        k.attr['abc'] = 10
        self.assertDictEqual(k.attr, {'abc': 10})
        self.check_unobserved(k)

    def test_dict_indexing_invalid_key(self):
        """ Dict [] invalid key. """

        k = Klass(attr={'abc': 10})
        with self.assertRaises(ValidationError):
            k.attr[10] = 20

    def test_dict_indexing_invalid_value(self):
        """ Dict [] invalid value. """

        k = Klass(attr={'abc': 10})
        with self.assertRaises(ValidationError):
            k.attr['abc'] = 'xyz'

    def test_dict_deleting_valid(self):
        """ Dict del valid key. """

        k = Klass(attr={'abc': 10})
        del k.attr['abc']
        self.assertDictEqual(k.attr, {})
        self.check_observed(k, {}, {'abc': 10})

    def test_dict_deleting_invalid(self):
        """ Dict del invalid key. """

        k = Klass(attr={'abc': 10})
        with self.assertRaises(ValidationError):
            del k.attr[10]


if __name__ == '__main__':
    unittest.main()
