# Copyright (c) 2012 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import (Any, Bool, Callable, Float, Int, Model, Str, Trigger,
        ValidationError)


class TestTypes(unittest.TestCase):
    """ This tests the basic types. """

    def test_del_model_attribute(self):
        """ Cannot delete model attributes. """

        class Klass(Model):
            attr = Any()

        k = Klass()

        with self.assertRaises(AttributeError):
            del k.attr

    def test_metadata(self):
        """ Type meta-data. """

        class Klass(Model):
            attr = Any(test='tested')

        self.assertEquals(Klass.attr.metadata['test'], 'tested')

    def test_default_value(self):
        """ Default values. """

        class Klass1(Model):
            attr = Any('explicit')

        class Klass2(Klass1):
            @Klass1.attr.default
            def attr(self):
                return 'default'

        # Verify default values.
        k1 = Klass1()
        self.assertEqual(k1.attr, 'explicit')

        k2 = Klass2()
        self.assertEqual(k2.attr, 'default')

        # Verify overriden default values.
        k2 = Klass2(attr='__init__')
        self.assertEqual(k2.attr, '__init__')

    def test_subclass_default_value(self):
        """ Default values in sub-classes. """

        class Klass1(Model):
            attr = Any('explicit')

        class Klass2(Klass1):
            attr = 'new_explicit'

        # Verify default values.
        k2 = Klass2()
        self.assertEqual(k2.attr, 'new_explicit')

    def test_subclass_overridden_value(self):
        """ Default values in sub-classes that are overridden. """

        class Klass1(Model):
            attr = Any('explicit')

        class Klass2(Klass1):
            attr = 'new_explicit'

        # Verify overridden values.
        k2 = Klass2(attr='init')
        self.assertEqual(k2.attr, 'init')

    def test_invalid_default_value(self):
        """ Invalid default values. """

        class Klass(Model):
            attr = Str()

            @attr.default
            def attr(self):
                return False

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr

    def test_Any(self):
        """ The Any type. """

        self._test_type(attribute_type=Any, default_value=None,
                non_default_value='non-default')

    def test_Bool(self):
        """ The Bool type. """

        class FakeBool:
            def __bool__(self):
                return True

        self._test_type(attribute_type=Bool, default_value=False,
                non_default_value=FakeBool(), normalised_value=True)

    def test_Callable(self):
        """ The Callable type. """

        self._test_validating_type(attribute_type=Callable, default_value=None,
                non_default_value=int, invalid_value=True)

    def test_Int(self):
        """ The Int type. """

        self._test_validating_type(attribute_type=Int, default_value=0,
                non_default_value=10, invalid_value='invalid')

    def test_Float(self):
        """ The Float type. """

        self._test_validating_type(attribute_type=Float, default_value=0.0,
                non_default_value=10.0, invalid_value='invalid')

    def test_Trigger(self):
        """ The Trigger type. """

        class Klass(Model):
            attr = Trigger()

        k = Klass()

        # Verify that a signal cannot be read.
        with self.assertRaises(AttributeError):
            k.attr

    def test_Str(self):
        """ The Str type. """

        self._test_validating_type(attribute_type=Str, default_value='',
                non_default_value='non_default', invalid_value=False)

    def test_Float_None(self):
        """ The Float type allowing None. """

        self._test_allow_none(Float)

    def test_Int_None(self):
        """ The Int type allowing None. """

        self._test_allow_none(Int)

    def test_Str_None(self):
        """ The Str type allowing None. """

        self._test_allow_none(Str)

    def _test_allow_none(self, attribute_type):
        """ Test a type's handling of allow_none. """

        class Klass(Model):
            attr = attribute_type(allow_none=True)

        k = Klass(attr=None)

        self.assertIsNone(k.attr)

    def _test_validating_type(self, attribute_type, default_value,
            non_default_value, invalid_value):
        """ Test a type that can have invalid values. """

        # Test the valid values.
        self._test_type(attribute_type, default_value, non_default_value)

        # Create a valid test class.
        class Klass(Model):
            attr = attribute_type()

        # Verify an invalid value.
        with self.assertRaises(ValidationError):
            Klass(attr=invalid_value)

        # Verify rebinding to an invalid value.
        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = invalid_value

    def _test_type(self, attribute_type, default_value, non_default_value,
            normalised_value=None):
        """ Test a type that cannot have invalid values. """

        # Use the non-default value if there is no normalised value.
        if normalised_value is None:
            normalised_value = non_default_value

        # Create a valid test class.
        class Klass(Model):
            attr_default = attribute_type()
            attr = attribute_type(non_default_value)

        # Verify the default default value and an explicit default value.
        k = Klass()
        self.assertEqual(k.attr_default, default_value)
        self.assertEqual(k.attr, normalised_value)

        # Verify a valid value.
        k = Klass(attr=non_default_value)
        self.assertEqual(k.attr, normalised_value)

        # Verify rebinding to a valid value.
        k = Klass()
        k.attr = non_default_value
        self.assertEqual(k.attr, normalised_value)


if __name__ == '__main__':
    unittest.main()
