# Copyright (c) 2012 Riverbank Computing Limited.


if __name__ == '__main__':
    import sys
    from PyQt4.QtGui import QApplication

    app = QApplication(sys.argv)


try:
    import unittest2 as unittest
except ImportError:
    import unittest

import os

from dip.automate import Robot
from dip.io import IoManager
from dip.io.codecs.xml import XmlCodec
from dip.io.storage.filesystem import FilesystemStorageFactory
from dip.model import (Bool, Enum, Float, implements, Int, Model, observe, Str,
        Trigger)
from dip.ui import (CheckBox, ComboBox, FilesystemLocationEditor, FloatSpinBox,
        LineEditor, OptionList, PushButton, RadioButtons, SpinBox,
        StorageLocationEditor, TextEditor, ToolButton)


class TestEditors(unittest.TestCase):
    """ This tests editors based on a model. """

    def test_CheckBox(self):
        """ Test a CheckBox editor. """

        class TestModel(Model):
            attr = Bool(False)

        model = TestModel()
        view = CheckBox('attr', id='Editors.CheckBox')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.CheckBox', 'set', True)
        self.assertTrue(model.attr)

    def test_ComboBox(self):
        """ Test a ComboBox editor. """

        class TestModel(Model):
            attr = Enum('yes', 'no')

        model = TestModel()
        view = ComboBox('attr', id='Editors.ComboBox')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.ComboBox', 'set', 'no')
        self.assertEqual(model.attr, 'no')

    def test_DirectorySelector(self):
        """ Test a FilesystemLocationEditor selecting a directory. """

        class TestModel(Model):
            attr = Str("/Replace/this")

        model = TestModel()
        view = FilesystemLocationEditor('attr', mode='directory',
                id='Editors.DirectorySelector')
        ui = view(model=model)
        ui.show()

        # Get a known valid directory name.
        valid = os.path.dirname(__file__)

        Robot.simulate('Editors.DirectorySelector', 'set', valid)
        self.assertEqual(model.attr, valid)

    def test_FileSelector(self):
        """ Test a FilesystemLocationEditor selecting a file. """

        class TestModel(Model):
            attr = Str("/Replace/this")

        model = TestModel()
        view = FilesystemLocationEditor('attr', id='Editors.FileSelector')
        ui = view(model=model)
        ui.show()

        # Get a known valid file name.
        valid = __file__

        Robot.simulate('Editors.FileSelector', 'set', valid)
        self.assertEqual(model.attr, valid)

    def test_StorageLocationEditor(self):
        """ Test a StorageLocationEditor selecting a file. """

        class TestModel(Model):
            attr = Str("/Replace/this")

        IoManager.codecs.append(XmlCodec(format='test.format'))
        IoManager.storage_factories.append(FilesystemStorageFactory())

        model = TestModel()
        view = StorageLocationEditor('attr', format='test.format',
                id='Editors.StorageLocationEditor')
        ui = view(model=model)
        ui.show()

        # Get a known valid file name.
        valid = __file__

        Robot.simulate('Editors.StorageLocationEditor', 'set', valid)
        self.assertEqual(model.attr, valid)

    def test_FloatSpinBox(self):
        """ Test a FloatSpinBox editor. """

        class TestModel(Model):
            attr = Float(10.0)

        model = TestModel()
        view = FloatSpinBox('attr', id='Editors.FloatSpinBox')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.FloatSpinBox', 'set', 20.0)
        self.assertEqual(model.attr, 20.0)

    def test_LineEditor(self):
        """ Test a LineEditor editor. """

        class TestModel(Model):
            attr = Str("Replace this...")

        model = TestModel()
        view = LineEditor('attr', id='Editors.LineEditor')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.LineEditor', 'set', "...with this")
        self.assertEqual(model.attr, "...with this")

    @unittest.skip("fix automation so the right QListWidget adapter is chosen")
    def test_OptionList(self):
        """ Test an OptionList editor. """

        class TestModel(Model):
            attr = Str("First option")

        model = TestModel()
        view = OptionList('attr', options=["First option", "Second option"],
                id='Editors.OptionList')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.OptionList', 'select', 1)
        self.assertEqual(model.attr, "Second option")

    def test_PushButton(self):
        """ Test a PushButton editor. """

        class TestModel(Model):
            attr = Trigger()

            @observe('attr')
            def __attr_changed(self, change):
                self.observed = change

        model = TestModel()
        view = PushButton('attr', id='Editors.PushButton')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.PushButton', 'trigger')
        self.assertTrue(model.observed)

    def test_RadioButtons(self):
        """ Test a RadioButtons editor. """

        class TestModel(Model):
            attr = Enum('yes', 'no')

        model = TestModel()
        view = RadioButtons('attr', id='Editors.RadioButtons')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.RadioButtons', 'set', 'no')
        self.assertEqual(model.attr, 'no')

    def test_SpinBox(self):
        """ Test a SpinBox editor. """

        class TestModel(Model):
            attr = Int(10)

        model = TestModel()
        view = SpinBox('attr', id='Editors.SpinBox')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.SpinBox', 'set', 20)
        self.assertEqual(model.attr, 20)

    def test_TextEditor(self):
        """ Test a TextEditor editor. """

        class TestModel(Model):
            attr = Str("Replace this...")

        model = TestModel()
        view = TextEditor('attr', id='Editors.TextEditor')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.TextEditor', 'set', "...with this")
        self.assertEqual(model.attr, "...with this")

    def test_ToolButton(self):
        """ Test a ToolButton editor. """

        class TestModel(Model):
            attr = Trigger()

            @observe('attr')
            def __attr_changed(self, change):
                self.observed = change

        model = TestModel()
        view = ToolButton('attr', id='Editors.ToolButton')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Editors.ToolButton', 'trigger')
        self.assertTrue(model.observed)


if __name__ == '__main__':
    unittest.main()
