# Copyright (c) 2011 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from PyQt4.QtCore import  QRegExp
from PyQt4.QtGui import QApplication
from PyQt4.QtTest import QTest

from dip.automate import Robot
from dip.model import (Bool, Enum, Float, Instance, Int, List, Model, observe,
        Str, Trigger)
from dip.ui import Form


if __name__ == '__main__':
    import sys

    app = QApplication(sys.argv)


class TestModel(unittest.TestCase):
    """ This tests views based on a model. """

    def test_nested_model(self):
        """ Test a nested model. """

        class Klass1(Model):
            attr = Str()

        class Klass2(Model):
            inst = Instance(Klass1)

        k1 = Klass1()
        k2 = Klass2(inst=k1)

        view = Form('inst.attr', id='Model.nested_model')
        ui = view(model=k2)
        ui.show()

        Robot.simulate('Model.nested_model:inst.attr', 'set', 'non-default')
        self.assertEqual(k2.inst.attr, 'non-default')

    def test_Bool(self):
        """ Test Bool attributes of a model. """

        class TestModel(Model):
            attr = Bool(False)

        model = TestModel()
        view = Form(id='Model.Bool')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Bool:attr', 'set', True)
        self.assertIs(model.attr, True)

    def test_Int(self):
        """ Test Int attributes of a model. """

        class TestModel(Model):
            attr = Int(10)

        model = TestModel()
        view = Form(id='Model.Int')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Int:attr', 'set', 20)
        self.assertEqual(model.attr, 20)

    def test_Int_minmax(self):
        """ Test Int attributes of a model with minimum and maximum
        constraints.
        """

        class TestModel(Model):
            attr = Int(10, minimum=5, maximum=15)

        model = TestModel()
        view = Form(id='Model.Int_minmax')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Int_minmax:attr', 'set', 20)
        self.assertEqual(model.attr, 10)

        Robot.simulate('Model.Int_minmax:attr', 'set', 0)
        self.assertEqual(model.attr, 10)

    def test_Float(self):
        """ Test Float attributes of a model. """

        class TestModel(Model):
            attr = Float(10.0)

        model = TestModel()
        view = Form(id='Model.Float')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Float:attr', 'set', 20)
        self.assertEqual(model.attr, 20.0)

    def test_Float_minmax(self):
        """ Test Float attributes of a model with minimum and maximum
        constraints.
        """

        class TestModel(Model):
            attr = Float(10.0, minimum=5.0, maximum=15.0)

        model = TestModel()
        view = Form(id='Model.Float_minmax')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Float_minmax:attr', 'set', 20)
        self.assertEqual(model.attr, 10.0)

        Robot.simulate('Model.Float_minmax:attr', 'set', 0)
        self.assertEqual(model.attr, 10.0)

    def test_Str(self):
        """ Test Str attributes of a model. """

        class TestModel(Model):
            attr = Str('default')

        model = TestModel()
        view = Form(id='Model.Str')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Str:attr', 'set', 'non-default')
        self.assertEqual(model.attr, 'non-default')

    def test_Str_validator(self):
        """ Test Str attributes of a model with a validator. """

        class TestModel(Model):
            attr = Str('default', required='stripped')

        model = TestModel()
        view = Form(id='Model.Str_validator')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Str_validator:attr', 'set', ' non default ')
        self.assertEqual(model.attr, 'default')

        Robot.simulate('Model.Str_validator:attr', 'set', 'non default')
        self.assertEqual(model.attr, 'non default')

    def test_Trigger(self):
        """ Test Trigger attributes of a model. """

        class TestModel(Model):
            attr = Trigger()

            def __init__(self):
                self.observed = False
                observe('attr', self, self.__observe)

            def __observe(self):
                self.observed = True

        model = TestModel()
        view = Form(id='Model.Trigger')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Trigger:attr', 'trigger')
        self.assertTrue(model.observed)

    def test_Enum(self):
        """ Test Enum attributes of a model. """

        class TestModel(Model):
            attr = Enum('foo', 'bar', 'baz')

        model = TestModel()
        view = Form(id='Model.Enum')
        ui = view(model=model)
        ui.show()

        Robot.simulate('Model.Enum:attr', 'set', 'baz')
        self.assertEqual(model.attr, 'baz')

    def test_List(self):
        """ Test List attributes of a model. """

        class TestModel(Model):
            attr = List()

        model = TestModel(attr=['first', 'second'])
        view = Form(id='Model.List')
        ui = view(model=model)
        ui.show()
        QTest.qWaitForWindowShown(ui)
        QApplication.setActiveWindow(ui)

        Robot.simulate('Model.List:attr', 'update', '2nd', 1)
        self.assertEqual(model.attr, ['first', '2nd'])


if __name__ == '__main__':
    unittest.main()
